@@ -4,9 +4,8 @@ import uuid
 from dataclasses import dataclass
 from datetime import timedelta
 from typing import Callable, Optional, Union
-from google.protobuf.struct_pb2 import Struct
 
-from omotes_sdk.internal.common.broker_interface import BrokerInterface
+from omotes_sdk.internal.common.broker_interface import BrokerInterface, AMQPQueueType
 from omotes_sdk.config import RabbitMQConfig
 from omotes_sdk_protocol.job_pb2 import (
     JobResult,
@@ -16,6 +15,8 @@ from omotes_sdk_protocol.job_pb2 import (
     JobCancel,
 )
 from omotes_sdk_protocol.workflow_pb2 import AvailableWorkflows, RequestAvailableWorkflows
+
+from omotes_sdk.internal.worker.params_dict import convert_params_dict_to_struct
 from omotes_sdk.job import Job
 from omotes_sdk.queue_names import OmotesQueueNames
 from omotes_sdk.types import ParamsDict
@@ -94,30 +95,56 @@ class OmotesInterface:
     """All available workflow types."""
     _workflow_config_received: threading.Event
     """Event triggered when workflow configuration is received."""
+    client_id: str
+    """Identifier of this SDK instance. Should be unique from other SDKs."""
+    timeout_on_initial_workflow_definitions: timedelta
+    """How long the SDK should wait for the first reply when requesting the current workflow
+    definitions from the orchestrator."""
 
     def __init__(
         self,
         rabbitmq_config: RabbitMQConfig,
+        client_id: str,
+        timeout_on_initial_workflow_definitions: timedelta = timedelta(minutes=1),
     ):
         """Create the OMOTES interface.
 
         NOTE: Needs to be started separately.
 
         :param rabbitmq_config: RabbitMQ configuration how to connect to OMOTES.
+        :param client_id: Identifier of this SDK instance. Should be unique from other SDKs.
+        :param timeout_on_initial_workflow_definitions: How long the SDK should wait for the first
+            reply when requesting the current workflow definitions from the orchestrator.
         """
         self.broker_if = BrokerInterface(rabbitmq_config)
         self.workflow_type_manager = None
         self._workflow_config_received = threading.Event()
+        self.client_id = client_id
+        self.timeout_on_initial_workflow_definitions = timeout_on_initial_workflow_definitions
 
     def start(self) -> None:
         """Start any other interfaces and request available workflows."""
         self.broker_if.start()
+        self.broker_if.declare_exchange(OmotesQueueNames.omotes_exchange_name())
+        # Declare job submissions queue just in case the orchestrator isn't life yet so we do not
+        # lose work.
+        self.broker_if.declare_queue(
+            queue_name=OmotesQueueNames.job_submission_queue_name(),
+            queue_type=AMQPQueueType.DURABLE,
+            exchange_name=OmotesQueueNames.omotes_exchange_name(),
+        )
         self.connect_to_available_workflows_updates()
         self.request_available_workflows()
 
-        while not self._workflow_config_received.is_set():
-            logger.info("Waiting for workflow definitions to be received from the orchestrator...")
-            self._workflow_config_received.wait(timeout=5)
+        logger.info("Waiting for workflow definitions to be received from the orchestrator...")
+        if not self._workflow_config_received.wait(
+            timeout=self.timeout_on_initial_workflow_definitions.total_seconds()
+        ):
+            raise RuntimeError(
+                "The orchestrator did not send the available workflows within "
+                "the expected time. Is the orchestrator online and connected "
+                "to the broker?"
+            )
 
     def stop(self) -> None:
         """Stop any other interfaces."""
@@ -128,11 +155,14 @@ class OmotesInterface:
 
         :param job: Job to disconnect from.
         """
-        self.broker_if.remove_queue_next_message_subscription(
-            OmotesQueueNames.job_results_queue_name(job)
-        )
-        self.broker_if.remove_queue_subscription(OmotesQueueNames.job_progress_queue_name(job))
-        self.broker_if.remove_queue_subscription(OmotesQueueNames.job_status_queue_name(job))
+        self.broker_if.remove_queue_subscription(OmotesQueueNames.job_results_queue_name(job.id))
+        self.broker_if.remove_queue_subscription(OmotesQueueNames.job_progress_queue_name(job.id))
+        self.broker_if.remove_queue_subscription(OmotesQueueNames.job_status_queue_name(job.id))
+
+    def _autodelete_progres_status_queues_on_result(self, job: Job) -> None:
+        """Disconnect and delete the progress and status queues for some job."""
+        self.broker_if.remove_queue_subscription(OmotesQueueNames.job_progress_queue_name(job.id))
+        self.broker_if.remove_queue_subscription(OmotesQueueNames.job_status_queue_name(job.id))
 
     def connect_to_submitted_job(
         self,
@@ -152,11 +182,12 @@ class OmotesInterface:
         :param callback_on_progress_update: Called when there is a progress update for the job.
         :param callback_on_status_update: Called when there is a status update for the job.
         :param auto_disconnect_on_result: Remove/disconnect from all queues pertaining to this job
-        once the result is received and handled without exceptions through `callback_on_finished`.
+            once the result is received and handled without exceptions through
+            `callback_on_finished`.
         """
         if auto_disconnect_on_result:
             logger.info("Connecting to update for job %s with auto disconnect on result", job.id)
-            auto_disconnect_handler = self.disconnect_from_submitted_job
+            auto_disconnect_handler = self._autodelete_progres_status_queues_on_result
         else:
             logger.info("Connecting to update for job %s and expect manual disconnect", job.id)
             auto_disconnect_handler = None
@@ -169,21 +200,26 @@ class OmotesInterface:
             auto_disconnect_handler,
         )
 
-        self.broker_if.receive_next_message(
-            queue_name=OmotesQueueNames.job_results_queue_name(job),
-            timeout=None,
+        self.broker_if.declare_queue_and_add_subscription(
+            queue_name=OmotesQueueNames.job_results_queue_name(job.id),
             callback_on_message=callback_handler.callback_on_finished_wrapped,
-            callback_on_no_message=None,
+            queue_type=AMQPQueueType.DURABLE,
+            exchange_name=OmotesQueueNames.omotes_exchange_name(),
+            delete_after_messages=1,
         )
         if callback_on_progress_update:
-            self.broker_if.add_queue_subscription(
-                queue_name=OmotesQueueNames.job_progress_queue_name(job),
+            self.broker_if.declare_queue_and_add_subscription(
+                queue_name=OmotesQueueNames.job_progress_queue_name(job.id),
                 callback_on_message=callback_handler.callback_on_progress_update_wrapped,
+                queue_type=AMQPQueueType.DURABLE,
+                exchange_name=OmotesQueueNames.omotes_exchange_name(),
             )
         if callback_on_status_update:
-            self.broker_if.add_queue_subscription(
-                queue_name=OmotesQueueNames.job_status_queue_name(job),
+            self.broker_if.declare_queue_and_add_subscription(
+                queue_name=OmotesQueueNames.job_status_queue_name(job.id),
                 callback_on_message=callback_handler.callback_on_status_update_wrapped,
+                queue_type=AMQPQueueType.DURABLE,
+                exchange_name=OmotesQueueNames.omotes_exchange_name(),
             )
 
     def submit_job(
@@ -205,6 +241,7 @@ class OmotesInterface:
             str, Union[Struct, ListValue, str, float, bool, None, Mapping[str, Any], Sequence]
         :param workflow_type: Type of the workflow to start.
         :param job_timeout: How long the job may take before it is considered to be timeout.
+            If None is given, the job is immune from the periodic timeout job cancellation task.
         :param callback_on_finished: Callback which is called with the job result once the job is
             done.
         :param callback_on_progress_update: Callback which is called with any progress updates.
@@ -232,18 +269,21 @@ class OmotesInterface:
             auto_disconnect_on_result,
         )
 
-        timeout_ms = round(job_timeout.total_seconds() * 1000) if job_timeout else None
-        params_dict_struct = Struct()
-        params_dict_struct.update(params_dict)
+        if job_timeout is not None:
+            timeout_ms = round(job_timeout.total_seconds() * 1000)
+        else:
+            timeout_ms = None
+
         job_submission_msg = JobSubmission(
             uuid=str(job.id),
             timeout_ms=timeout_ms,
             workflow_type=workflow_type.workflow_type_name,
             esdl=esdl,
-            params_dict=params_dict_struct,
+            params_dict=convert_params_dict_to_struct(params_dict),
         )
         self.broker_if.send_message_to(
-            OmotesQueueNames.job_submission_queue_name(workflow_type),
+            exchange_name=OmotesQueueNames.omotes_exchange_name(),
+            routing_key=OmotesQueueNames.job_submission_queue_name(),
             message=job_submission_msg.SerializeToString(),
         )
         logger.debug("Done submitting job %s", job.id)
@@ -262,14 +302,19 @@ class OmotesInterface:
         logger.info("Cancelling job %s", job.id)
         cancel_msg = JobCancel(uuid=str(job.id))
         self.broker_if.send_message_to(
-            OmotesQueueNames.job_cancel_queue_name(), message=cancel_msg.SerializeToString()
+            exchange_name=OmotesQueueNames.omotes_exchange_name(),
+            routing_key=OmotesQueueNames.job_cancel_queue_name(),
+            message=cancel_msg.SerializeToString(),
         )
 
     def connect_to_available_workflows_updates(self) -> None:
         """Connect to updates of the available workflows."""
-        self.broker_if.add_queue_subscription(
-            queue_name=OmotesQueueNames.available_workflows_queue_name(),
+        self.broker_if.declare_queue_and_add_subscription(
+            queue_name=OmotesQueueNames.available_workflows_queue_name(self.client_id),
             callback_on_message=self.callback_on_update_available_workflows,
+            queue_type=AMQPQueueType.EXCLUSIVE,
+            bind_to_routing_key=OmotesQueueNames.available_workflows_routing_key(),
+            exchange_name=OmotesQueueNames.omotes_exchange_name(),
         )
 
     def callback_on_update_available_workflows(self, message: bytes) -> None:
@@ -287,8 +332,9 @@ class OmotesInterface:
         """Request the available workflows from the orchestrator."""
         request_available_workflows_pb = RequestAvailableWorkflows()
         self.broker_if.send_message_to(
-            OmotesQueueNames.request_available_workflows_queue_name(),
-            request_available_workflows_pb.SerializeToString(),
+            exchange_name=OmotesQueueNames.omotes_exchange_name(),
+            routing_key=OmotesQueueNames.request_available_workflows_queue_name(),
+            message=request_available_workflows_pb.SerializeToString(),
         )
 
     def get_workflow_type_manager(self) -> WorkflowTypeManager:
